"""Functions for converting and mapping between geographic types."""
import re
import warnings
from typing import Union, Iterable

import pandas as pd
import pkg_resources

COUNTY_CENSUS = pd.read_csv(
    pkg_resources.resource_filename(__name__, "geo_mappings/county_census.csv"), dtype=str)
MSA_CENSUS = pd.read_csv(
    pkg_resources.resource_filename(__name__, "geo_mappings/msa_census.csv"), dtype=str)
STATE_CENSUS = pd.read_csv(
    pkg_resources.resource_filename(__name__, "geo_mappings/state_census.csv"), dtype=str)

# Filter undesired rows from CSVs.
# They're not removed from the files to keep them identical to rda files.
STATE_CENSUS = STATE_CENSUS.loc[STATE_CENSUS.STATE != "0"]
# pad to 2 characters with leading 0s
STATE_CENSUS["STATE"] = STATE_CENSUS["STATE"].str.zfill(2)
# add 000 to the end to get a 5 digit code
STATE_CENSUS["STATE"] = STATE_CENSUS["STATE"].str.pad(width=5, fillchar="0", side="right")
# filter out micropolitan areas
MSA_CENSUS = MSA_CENSUS.loc[MSA_CENSUS.LSAD == "Metropolitan Statistical Area"]


def fips_to_name(code: Union[str, Iterable[str]],
                 ignore_case: bool = False,
                 fixed: bool = False,
                 ties_method: str = "first") -> list:
    """Look up county names by FIPS codes with regular expression support.

    Given an individual or list of FIPS codes or regular expressions, look up the corresponding
    county names.

    :param code: Individual or list of FIPS codes or regular expressions.
    :param ignore_case: Boolean for whether or not to be case insensitive in the regular expression.
      If ``fixed=True``, this argument is ignored. Defaults to ``False``.
    :param fixed: Conduct an exact case sensitive match with the input string.
      Defaults to ``False``.
    :param ties_method: Method for determining how to deal with multiple outputs for a given input.
      Must be one of ``"all"`` or ``"first"``. If ``"first"``, then only the first match for each
      code is returned. If ``"all"``, then all matches for each code are returned.
      Defaults to ``first``.
    :return: If ``ties_method="first"``, returns a list of the first value found for each input key.
      If ``ties_method="all"``, returns a list of dicts, one for each input, with keys
      corresponding to all matched input keys and values corresponding to the list of county names.
      The returned list will be the same length as the input, with ``None`` or ``{}`` if no values
      are found for ``ties_method="first"`` and ``ties_method="all"``, respectively.
    """
    return _lookup(code, COUNTY_CENSUS.FIPS, COUNTY_CENSUS.CTYNAME, ignore_case, fixed, ties_method)


def cbsa_to_name(code: Union[str, Iterable[str]],
                 ignore_case: bool = False,
                 fixed: bool = False,
                 ties_method: str = "first") -> list:
    """Look up MSA names by codes with regular expression support.

    Given an individual or list of FIPS codes or regular expressions, look up the corresponding
    MSA names.

    :param code: Individual or list of FIPS codes or regular expressions.
    :param ignore_case: Boolean for whether or not to be case insensitive in the regular expression.
      If ``fixed=True``, this argument is ignored. Defaults to ``False``.
    :param fixed: Conduct an exact case sensitive match with the input string.
      Defaults to ``False``.
    :param ties_method: Method for determining how to deal with multiple outputs for a given input.
      Must be one of ``"all"`` or ``"first"``. If ``"first"``, then only the first match for each
      code is returned. If ``"all"``, then all matches for each code are returned.
      Defaults to ``first``.
    :return: If ``ties_method="first"``, returns a list of the first value found for each input key.
      If ``ties_method="all"``, returns a list of dicts, one for each input, with keys
      corresponding to all matched input keys and values corresponding to the list of MSA names.
      The returned list will be the same length as the input, with ``None`` or ``{}`` if no values
      are found for ``ties_method="first"`` and ``ties_method="all"``, respectively.
    """
    return _lookup(code, MSA_CENSUS.CBSA, MSA_CENSUS.NAME, ignore_case, fixed, ties_method)


def abbr_to_name(abbr: Union[str, Iterable[str]],
                 ignore_case: bool = False,
                 fixed: bool = False,
                 ties_method: str = "first") -> list:
    """Look up state name by abbreviation with regular expression support.

    Given an individual or list of state abbreviations or regular expressions, look up the
    corresponding state names.

    :param abbr: Individual or list of state abbreviations or regular expressions.
    :param ignore_case: Boolean for whether or not to be case insensitive in the regular expression.
      If ``fixed=True``, this argument is ignored. Defaults to ``False``.
    :param fixed: Conduct an exact case sensitive match with the input string.
      Defaults to ``False``.
    :param ties_method: Method for determining how to deal with multiple outputs for a given input.
      Must be one of ``"all"`` or ``"first"``. If ``"first"``, then only the first match for each
      code is returned. If ``"all"``, then all matches for each code are returned.
      Defaults to ``first``.
    :return: If ``ties_method="first"``, returns a list of the first value found for each input key.
      If ``ties_method="all"``, returns a list of dicts, one for each input, with keys
      corresponding to all matched input keys and values corresponding to the list of state names.
      The returned list will be the same length as the input, with ``None`` or ``{}`` if no values
      are found for ``ties_method="first"`` and ``ties_method="all"``, respectively.
    """
    return _lookup(abbr, STATE_CENSUS.ABBR, STATE_CENSUS.NAME, ignore_case, fixed, ties_method)


def name_to_abbr(name: Union[str, Iterable[str]],
                 ignore_case: bool = False,
                 fixed: bool = False,
                 ties_method: str = "first") -> list:
    """Look up state abbreviation by name with regular expression support.

    Given an individual or list of state names or regular expressions, look up the
    corresponding state abbreviations.

    :param name: Individual or list of state names or regular expressions.
    :param ignore_case: Boolean for whether or not to be case insensitive in the regular expression.
      If ``fixed=True``, this argument is ignored. Defaults to ``False``.
    :param fixed: Conduct an exact case sensitive match with the input string.
      Defaults to ``False``.
    :param ties_method: Method for determining how to deal with multiple outputs for a given input.
      Must be one of ``"all"`` or ``"first"``. If ``"first"``, then only the first match for each
      code is returned. If ``"all"``, then all matches for each code are returned.
      Defaults to ``first``.
    :return: If ``ties_method="first"``, returns a list of the first value found for each input key.
      If ``ties_method="all"``, returns a list of dicts, one for each input, with keys
      corresponding to all matched input keys and values corresponding to the list of
      state abbreviations. The returned list will be the same length as the input, with ``None`` or
      ``{}`` if no values are found for ``ties_method="first"`` and ``ties_method="all"``,
      respectively.
    """
    return _lookup(name, STATE_CENSUS.NAME, STATE_CENSUS.ABBR, ignore_case, fixed, ties_method)


def fips_to_abbr(code: Union[str, Iterable[str]],
                 ignore_case: bool = False,
                 fixed: bool = False,
                 ties_method: str = "first") -> list:
    """Look up state abbreviation by FIPS codes with regular expression support.

    Given an individual or list of FIPS codes or regular expressions, look up the corresponding
    state abbreviation. FIPS codes can be the 2 digit code (``covidcast.fips_to_abbr("12")``) or
    the 2 digit code with 000 appended to the end (``covidcast.fips_to_abbr("12000")``.

    :param code: Individual or list of FIPS codes or regular expressions.
    :param ignore_case: Boolean for whether or not to be case insensitive in the regular expression.
      If ``fixed=True``, this argument is ignored. Defaults to ``False``.
    :param fixed: Conduct an exact case sensitive match with the input string.
      Defaults to ``False``.
    :param ties_method: Method for determining how to deal with multiple outputs for a given input.
      Must be one of ``"all"`` or ``"first"``. If ``"first"``, then only the first match for each
      code is returned. If ``"all"``, then all matches for each code are returned.
      Defaults to ``first``.
    :return: If ``ties_method="first"``, returns a list of the first value found for each input key.
      If ``ties_method="all"``, returns a list of dicts, one for each input, with keys
      corresponding to all matched input keys and values corresponding to the list of county names.
      The returned list will be the same length as the input, with ``None`` or ``{}`` if no values
      are found for ``ties_method="first"`` and ``ties_method="all"``, respectively.
    """
    return _lookup(code, STATE_CENSUS.STATE, STATE_CENSUS.ABBR, ignore_case, fixed, ties_method)


def name_to_cbsa(name: Union[str, Iterable[str]],
                 ignore_case: bool = False,
                 fixed: bool = False,
                 ties_method: str = "first",
                 state: str = None) -> list:
    """Look up MSA codes by names with regular expression support.

    Given an individual or list of names or regular expressions, look up the corresponding
    MSA codes.

    :param name: Individual or list of MSA names or regular expressions.
    :param ignore_case: Boolean for whether or not to be case insensitive in the regular expression.
      If ``fixed=True``, this argument is ignored. Defaults to ``False``.
    :param fixed: Conduct an exact case sensitive match with the input string.
      Defaults to ``False``.
    :param ties_method: Method for determining how to deal with multiple outputs for a given input.
      Must be one of ``"all"`` or ``"first"``. If ``"first"``, then only the first match for each
      code is returned. If ``"all"``, then all matches for each code are returned.
      Defaults to ``first``.
    :param state: 2 letter state code, case insensitive, to restrict results to.
    :return: If ``ties_method="first"``, returns a list of the first value found for each input key.
      If ``ties_method="all"``, returns a list of dicts, one for each input, with keys
      corresponding to all matched input keys and values corresponding to the list of MSA codes.
      The returned list will be the same length as the input, with ``None`` or ``{}`` if no values
      are found for ``ties_method="first"`` and ``ties_method="all"``, respectively.
    """
    if state:
        state = state.upper()
        df = MSA_CENSUS.loc[MSA_CENSUS.STATE == state]
    else:
        df = MSA_CENSUS
    return _lookup(name, df.NAME, df.CBSA, ignore_case, fixed, ties_method)


def abbr_to_fips(code: Union[str, Iterable[str]],
                 ignore_case: bool = False,
                 fixed: bool = False,
                 ties_method: str = "first") -> list:
    """Look up state FIPS codes by abbreviation with regular expression support.

    Given an individual or list of state abbreviations or regular expressions,
    look up the corresponding state FIPS codes. The returned codes are 5 digits: the
    2 digit state FIPS with 000 appended to the end.

    :param code: Individual or list of abbreviations or regular expressions.
    :param ignore_case: Boolean for whether or not to be case insensitive in the regular expression.
      If ``fixed=True``, this argument is ignored. Defaults to ``False``.
    :param fixed: Conduct an exact case sensitive match with the input string.
      Defaults to ``False``.
    :param ties_method: Method for determining how to deal with multiple outputs for a given input.
      Must be one of ``"all"`` or ``"first"``. If ``"first"``, then only the first match for each
      code is returned. If ``"all"``, then all matches for each code are returned.
      Defaults to ``first``.
    :return: If ``ties_method="first"``, returns a list of the first value found for each input key.
      If ``ties_method="all"``, returns a list of dicts, one for each input, with keys
      corresponding to all matched input keys and values corresponding to the list of county names.
      The returned list will be the same length as the input, with ``None`` or ``{}`` if no values
      are found for ``ties_method="first"`` and ``ties_method="all"``, respectively.
    """
    return _lookup(code, STATE_CENSUS.ABBR, STATE_CENSUS.STATE, ignore_case, fixed, ties_method)


def name_to_fips(name: Union[str, Iterable[str]],
                 ignore_case: bool = False,
                 fixed: bool = False,
                 ties_method: str = "first",
                 state: str = None) -> list:
    """Look up FIPS codes by county names with regular expression support.

    Given an individual or list of county names or regular expressions, look up the corresponding
    FIPS codes.

    :param name: Individual or list of county names or regular expressions.
    :param ignore_case: Boolean for whether or not to be case insensitive in the regular expression.
      If ``fixed=True``, this argument is ignored. Defaults to ``False``.
    :param fixed: Conduct an exact case sensitive match with the input string.
      Defaults to ``False``.
    :param ties_method: Method for determining how to deal with multiple outputs for a given input.
      Must be one of ``"all"`` or ``"first"``. If ``"first"``, then only the first match for each
      code is returned. If ``"all"``, then all matches for each code are returned.
      Defaults to ``first``.
    :param state: 2 letter state code, case insensitive, to restrict results to.
    :return: If ``ties_method="first"``, returns a list of the first value found for each input key.
      If ``ties_method="all"``, returns a list of dicts, one for each input, with keys
      corresponding to all matched input keys and values corresponding to the list of FIPS.
      The returned list will be the same length as the input, with ``None`` or ``{}`` if no values
      are found for ``ties_method="first"`` and ``ties_method="all"``, respectively.
    """
    if state:
        state = state.upper()
        df = COUNTY_CENSUS.loc[COUNTY_CENSUS.STNAME == abbr_to_name(state)[0]]
    else:
        df = COUNTY_CENSUS
    return _lookup(name, df.CTYNAME, df.FIPS, ignore_case, fixed, ties_method)


def _lookup(key: Union[str, Iterable[str]],
            keys: Iterable,
            values: Iterable,
            ignore_case: bool = False,
            fixed: bool = False,
            ties_method: str = "first") -> list:
    """Given an input, search for it in a list of keys and return the corresponding values.

    :param key: Individual or list of search strings or regular expression patterns.
    :param keys: List of keys to be searched
    :param values: List of values that correspond to keys.
    :param ignore_case: Boolean for whether or not to be case insensitive in the regular expression.
      If ``fixed=True``, this argument is ignored. Defaults to ``False``.
    :param fixed: Conduct an exact case sensitive match with the input string.
      Defaults to ``False``.
    :param ties_method: Method for determining how to deal with multiple outputs for a given input.
      Must be one of ``"all"`` or ``"first"``. If ``"first"``, then only the first match for each
      code is returned. If `"all"`, then all matches for each code are returned.
      Defaults to ``first``.
    :return: If ``ties_method="first"``, returns a list of the first value found for each input key.
      If ``ties_method="all"``, returns a list of dicts, one for each input, with keys
      corresponding to all matched input keys and values corresponding to the list of values.
      The returned list will be the same length as the input, with None or {} if no values are found
      for ``ties_method="first"`` and ``ties_method="all"``, respectively.
    """
    if ties_method not in ("first", "all"):
        raise ValueError("Invalid `ties_method`. Must be one of `first` or `all`.")
    key = [key] if isinstance(key, str) else key
    case = re.IGNORECASE if (ignore_case and not fixed) else 0
    output = []
    for i in key:
        result: dict = {}
        for k, v in zip(keys, values):
            if i == k if fixed else re.search(i, k, case):
                result[k] = result.get(k, []) + [v]
        output.append(result)
    if ties_method == "first":
        return _get_first_tie(output)
    return output


def _get_first_tie(dict_list: list) -> list:
    """Return a list with the first value for the first key for each of the input dicts.

    Needs to be Python 3.6+ for this to work, since earlier versions don't preserve insertion order.

    :param dict_list: List of str:list dicts.
    :return: list of the first key and first value for that key for each of the input dicts.
    """
    not_unique = False
    for d in dict_list:
        if len(d) > 1 or any(len(val) > 1 for val in d.values()):
            not_unique = True
    if not_unique:
        warnings.warn("Some inputs were not uniquely matched; returning only the first match "
                      "in each case. To return all matches, set `ties_method='all'`")
    # first entry of first value
    return [list(d.values())[0][0] if d else None for d in dict_list]



"""This contains the plotting and geo data management methods for the COVIDcast signals."""

import io
import warnings
from datetime import date, timedelta
from typing import Tuple, Any

import geopandas as gpd
import imageio
import numpy as np
import pandas as pd
import pkg_resources
from matplotlib import pyplot as plt
from matplotlib import figure, axes
from tqdm import tqdm

from .covidcast import _detect_metadata, _signal_metadata

SHAPEFILE_PATHS = {"county": "shapefiles/county/cb_2019_us_county_5m.shp",
                   "state": "shapefiles/state/cb_2019_us_state_5m.shp",
                   "msa": "shapefiles/msa/cb_2019_us_cbsa_5m.shp",
                   "hrr": "shapefiles/hrr/geo_export_ad86cff5-e5ed-432e-9ec2-2ce8732099ee.shp"}

STATE_ABBR_TO_FIPS = {"AL": "01", "MN": "27", "ME": "23", "WA": "53", "LA": "22", "PA": "42",
                      "MD": "24", "CO": "08", "TN": "47", "MI": "26", "FL": "12", "VA": "51",
                      "IN": "18", "AS": "60", "HI": "15", "AZ": "04", "MO": "29", "SC": "45",
                      "DC": "11", "NM": "35", "MA": "25", "OR": "41", "MS": "28", "WI": "55",
                      "PR": "72", "NH": "33", "NV": "32", "GA": "13", "KY": "21", "NE": "31",
                      "WY": "56", "AK": "02", "OK": "40", "GU": "66", "DE": "10", "IA": "19",
                      "CA": "06", "VI": "78", "OH": "39", "NY": "36", "CT": "09", "AR": "05",
                      "VT": "50", "MP": "69", "MT": "30", "RI": "44", "WV": "54", "IL": "17",
                      "TX": "48", "UT": "49", "ND": "38", "KS": "20", "SD": "46", "NC": "37",
                      "NJ": "34", "ID": "16"}

# states within the contiguous US
CONTIGUOUS_FIPS = {"01", "04", "05", "06", "08", "09", "10", "11", "12", "13", "16", "17", "18",
                   "19", "20", "21", "22", "23", "24", "25", "26", "27", "28", "29", "30", "31",
                   "32", "33", "34", "35", "36", "37", "38", "39", "40", "41", "42", "44", "45",
                   "46", "47", "48", "49", "50", "51", "53", "54", "55", "56"}


def plot(data: pd.DataFrame,
         time_value: date = None,
         plot_type: str = "choropleth",
         combine_megacounties: bool = True,
         **kwargs: Any) -> figure.Figure:
    """Given the output data frame of :py:func:`covidcast.signal`, plot a choropleth or bubble map.

    Projections used for plotting:

    - ESRI:102003 (USA Contiguous Albers Equal Area Conic) for the contiguous US
      and Puerto Rico
    - ESRI:102006 (Alaska Albers Equal Area Conic) for Alaska
    - ESRI:102007 (Hawaii Albers Equal Area Conic) for Hawaii
      (Hawaii Albers Equal Area Conic) for Hawaii

    For visual purposes, Alaska and Hawaii are moved the lower left corner of the contiguous US
    and Puerto Rico is moved closer to Florida.

    By default, choropleths use the `colormap
    <https://matplotlib.org/tutorials/colors/colormaps.html>`_
    ``YlOrRd``, with colors scaled between 0 and the signal's historical mean value + 3
    standard deviations. Custom arguments can be passed in as ``kwargs`` for
    customizability. These arguments will be passed to the GeoPandas ``plot``
    method; more information on these arguments can be found in `the GeoPandas
    documentation
    <https://geopandas.org/reference.html#geopandas.GeoDataFrame.plot>`_.

    Bubble maps use a purple bubble by default, with all values discretized into 8 bins between 0.1
    and the signal's historical mean value + 3 standard deviations. Values below 0 have no
    bubble but have the region displayed in white, and values above the mean + 3 std dev are binned
    into the highest bubble. Bubbles are scaled by area.

    :param data: Data frame of signal values, as returned from :py:func:`covidcast.signal`.
    :param time_value: If multiple days of data are present in ``data``, map only values from this
      day. Defaults to plotting the most recent day of data in ``data``.
    :param combine_megacounties: For each state, display all counties without a signal value as a
      single polygon with the megacounty value, as opposed to plotting all the county boundaries.
      Defaults to `True`.
    :param kwargs: Optional keyword arguments passed to ``GeoDataFrame.plot()``.
    :param plot_type: Type of plot to create. Either choropleth (default) or bubble map.
    :return: Matplotlib figure object.

    """
    if plot_type not in {"choropleth", "bubble"}:
        raise ValueError("`plot_type` must be 'choropleth' or 'bubble'.")
    data_source, signal, geo_type = _detect_metadata(data)  # pylint: disable=W0212
    meta = _signal_metadata(data_source, signal, geo_type)  # pylint: disable=W0212
    # use most recent date in data if none provided
    day_to_plot = time_value if time_value else max(data.time_value)
    day_data = data.loc[data.time_value == pd.to_datetime(day_to_plot), :]
    kwargs["vmax"] = kwargs.get("vmax", meta["mean_value"] + 3 * meta["stdev_value"])
    kwargs["figsize"] = kwargs.get("figsize", (12.8, 9.6))

    fig, ax = _plot_background_states(kwargs["figsize"])
    ax.set_title(f"{data_source}: {signal}, {day_to_plot.strftime('%Y-%m-%d')}")
    if plot_type == "choropleth":
        _plot_choro(ax, day_data, combine_megacounties, **kwargs)
    else:
        _plot_bubble(ax, day_data, geo_type, **kwargs)
    return fig


def plot_choropleth(data: pd.DataFrame,
                    time_value: date = None,
                    combine_megacounties: bool = True,
                    **kwargs: Any) -> figure.Figure:
    """Plot choropleths for a signal. This method is deprecated and has been generalized to plot().

    :param data: Data frame of signal values, as returned from :py:func:`covidcast.signal`.
    :param time_value: If multiple days of data are present in ``data``, map only values from this
      day. Defaults to plotting the most recent day of data in ``data``.
    :param kwargs: Optional keyword arguments passed to ``GeoDataFrame.plot()``.
    :return: Matplotlib figure object.
    """
    warnings.warn("Function `plot_choropleth` is deprecated. Use `plot()` instead.")
    return plot(data, time_value, "choropleth", combine_megacounties, **kwargs)


def get_geo_df(data: pd.DataFrame,
               geo_value_col: str = "geo_value",
               geo_type_col: str = "geo_type",
               join_type: str = "right",
               combine_megacounties: bool = False) -> gpd.GeoDataFrame:
    """Augment a :py:func:`covidcast.signal` data frame with the shape of each geography.

    This method takes in a pandas DataFrame object and returns a GeoDataFrame
    object from the `GeoPandas package <https://geopandas.org/>`_. The
    GeoDataFrame will contain the geographic shape corresponding to every row in
    its ``geometry`` column; for example, a data frame of county-level signal
    observations will be returned with the shape of every county.

    After detecting the geography type (state, county, HRR, and MSA are currently
    supported) of the input, this function builds a GeoDataFrame that contains
    state and geometry information from the Census or CMS for that geography type. By
    default, it will take the signal data (left side) and geo data (right side)
    and right join them, so all states/counties will always be present
    regardless of whether ``data`` contains values for those locations. ``left``,
    ``outer``, and ``inner`` joins are also supported and can be selected with
    the ``join_type`` argument.

    If ``combine_megacounties=False`` (default) all counties without a signal value will be
    given the value of the megacounty if present. If ``combine_megacounties=True``, a left join
    will be conducted and the megacounty rows will be given a polygon of the union of all
    constituent counties without a value. Other joins will not use megacounties.
    See the `geographic coding documentation
    <https://cmu-delphi.github.io/delphi-epidata/api/covidcast_geography.html>`_
    for information about megacounties.

    By default, this function identifies the geography for each row of the input
    data frame using its ``geo_value`` column, matching data frames returned by
    :py:func:`covidcast.signal`, but the ``geo_value_col`` and ``geo_type_col``
    arguments can be provided to match geographies for data frames with
    different column names.

    Geographic data is sourced from 1:5,000,000-scale shapefiles from the `2019
    US Census Cartographic Boundary Files
    <https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.html>`_
    and the `CMS Data Website <https://data.cms.gov/widgets/ia25-mrsk>`_.

    :param data: DataFrame of values and geographies.
    :param geo_value_col: Name of column containing values of interest.
    :param geo_type_col: Name of column containing geography type.
    :param join_type: Type of join to do between input data (left side) and geo data (right side).
      Must be one of `right` (default), `left`, `outer`, or `inner`.
    :param combine_megacounties: For each state, return all counties without a signal value as a
      single row and polygon with the megacounty value. Defaults to `False`.
    :return: GeoDataFrame containing all columns from the input ``data``, along
      with a ``geometry`` column (containing a polygon) and a ``state_fips``
      column (a two-digit FIPS code identifying the US state containing this
      geography). For MSAs that span multiple states, the first state in the MSA name is provided.
      The geometry is given in the GCS NAD83 coordinate system for states, counties, and MSAs, and
      WGS84 for HRRs.

    """
    if join_type == "right" and any(data[geo_value_col].duplicated()):
        raise ValueError("join_type `right` is incompatible with duplicate values in a "
                         "given region. Use `left` or ensure your input data is a single signal for"
                         " a single date and geography type. ")
    geo_type = _detect_metadata(data, geo_type_col)[2]  # pylint: disable=W0212
    if geo_type not in ["state", "county", "msa", "hrr"]:
        raise ValueError("Unsupported geography type; "
                         "only `state`, `county`, `hrr`, and `msa` supported.")

    shapefile_path = pkg_resources.resource_filename(__name__, SHAPEFILE_PATHS[geo_type])
    geo_info = gpd.read_file(shapefile_path)

    if geo_type == "state":
        output = _join_state_geo_df(data, geo_value_col, geo_info, join_type)
    elif geo_type == "msa":
        output = _join_msa_geo_df(data, geo_value_col, geo_info, join_type)
    elif geo_type == "hrr":
        geo_info["geometry"] = geo_info["geometry"].translate(0, -0.185)  # fix projection shift bug
        output = _join_hrr_geo_df(data, geo_value_col, geo_info, join_type)
    else:  # geo_type must be "county"
        output = _join_county_geo_df(data, geo_value_col, geo_info, join_type, combine_megacounties)
    return output


def animate(data: pd.DataFrame, filepath: str, fps: int = 3, dpi: int = 150, **kwargs: Any) -> None:
    """Generate an animated video file of a signal over time.

    Given a signal DataFrame, generates the choropleth for each day to form an animation of the
    signal. Accepts arguments for video parameters as well as optional plotting arguments.
    Supported output formats are listed in the
    `imageio ffmpeg documentation <https://imageio.readthedocs.io/en/stable/format_ffmpeg.html>`_.

    :param data: DataFrame for a single signal over time.
    :param filepath: Path where video will be saved. Filename must contain supported extension.
    :param fps: Frame rate in frames per second for animation. Defaults to 3.
    :param dpi: Dots per inch for output video. Defaults to 150 on a 12.8x9.6 figure (1920x1440).
    :param kwargs: Optional keyword arguments passed to :py:func:`covidcast.plot`.
    :return: None
    """
    # probesize is set to avoid warning by ffmpeg on frame rate up to 4k resolution.
    writer = imageio.get_writer(filepath, fps=fps, input_params=["-probesize", "75M"])
    num_days = (max(data.time_value) - min(data.time_value)).days
    day_list = [min(data.time_value) + timedelta(days=x) for x in range(num_days+1)]
    for d in tqdm(day_list):
        buf = io.BytesIO()
        plot(data, time_value=d, **kwargs)
        plt.savefig(buf, dpi=dpi)
        plt.close()
        buf.seek(0)
        writer.append_data(imageio.imread(buf))
    writer.close()


def _plot_choro(ax: axes.Axes,
                data: gpd.GeoDataFrame,
                combine_megacounties: bool,
                **kwargs: Any) -> None:
    """Generate a choropleth map on a given Figure/Axes from a GeoDataFrame.

    :param ax: Matplotlib axes to plot on.
    :param data: GeoDataFrame with information to plot.
    :param kwargs: Optional keyword arguments passed to ``GeoDataFrame.plot()``.
    :return: Matplotlib axes with the plot added.
    """
    kwargs["vmin"] = kwargs.get("vmin", 0)
    kwargs["cmap"] = kwargs.get("cmap", "YlOrRd")
    data_w_geo = get_geo_df(data, combine_megacounties=combine_megacounties)
    for shape in _project_and_transform(data_w_geo):
        if not shape.empty:
            shape.plot(column="value", ax=ax, **kwargs)
    sm = plt.cm.ScalarMappable(cmap=kwargs["cmap"],
                               norm=plt.Normalize(vmin=kwargs["vmin"], vmax=kwargs["vmax"]))
    # this is to remove the set_array error that occurs on some platforms
    sm._A = []  # pylint: disable=W0212
    plt.colorbar(sm, ticks=np.linspace(kwargs["vmin"], kwargs["vmax"], 8), ax=ax,
                 orientation="horizontal", fraction=0.045, pad=0.04, format="%.2f")


def _plot_bubble(ax: axes.Axes, data: gpd.GeoDataFrame, geo_type: str, **kwargs: Any) -> None:
    """Generate a bubble map on a given Figure/Axes from a GeoDataFrame.

    The maximum bubble size is set to the figure area / 1.5, with a x3 multiplier if the geo_type
    is ``state``.

    :param ax: Matplotlib axes to plot on.
    :param data: GeoDataFrame with information to plot.
    :param kwargs: Optional keyword arguments passed to ``GeoDataFrame.plot()``.
    :return: Matplotlib axes with the plot added.
    """
    kwargs["vmin"] = kwargs.get("vmin", 0.1)
    kwargs["color"] = kwargs.get("color", "purple")
    kwargs["alpha"] = kwargs.get("alpha", 0.5)
    data_w_geo = get_geo_df(data, join_type="inner")
    label_bins = np.linspace(kwargs["vmin"], kwargs["vmax"], 8)  # set bin labels
    value_bins = list(label_bins) + [np.inf]  # set ranges for bins by adding +inf for largest bin
    # set max bubble size proportional to figure size, with a multiplier for state plots
    state_multiple = 3 if geo_type == "state" else 1
    bubble_scale = np.prod(kwargs["figsize"]) / 1.5 / kwargs["vmax"] * state_multiple
    # discretize data and scale labels to correct sizes
    data_w_geo["binval"] = pd.cut(data_w_geo.value, labels=label_bins, bins=value_bins, right=False)
    data_w_geo["binval"] = data_w_geo.binval.astype(float) * bubble_scale
    for shape in _project_and_transform(data_w_geo):
        if not shape.empty and not shape.binval.isnull().values.all():
            shape.plot(color="1", ax=ax, legend=True, edgecolor="0.8", linewidth=0.5)
            shape["geometry"] = shape["geometry"].centroid  # plot bubbles at each polgyon centroid
            shape.plot(markersize="binval", color=kwargs["color"], ax=ax, alpha=kwargs["alpha"])
    # to generate the legend, need to plot the reference points as scatter plots off the map
    for b in label_bins:
        ax.scatter([1e10], [1e10], color=kwargs["color"], alpha=kwargs["alpha"],
                   s=b*bubble_scale, label=round(b, 2))
    ax.legend(frameon=False, ncol=8, loc="lower center", bbox_to_anchor=(0.5, -0.1))


def _plot_background_states(figsize: tuple) -> tuple:
    """Plot US states in light grey as the background for other plots.

    :param figsize: Dimensions of plot.
    :return: Matplotlib figure and axes.
    """
    fig, ax = plt.subplots(1, figsize=figsize)
    ax.axis("off")
    state_shapefile_path = pkg_resources.resource_filename(__name__, SHAPEFILE_PATHS["state"])
    state = gpd.read_file(state_shapefile_path)
    for state in _project_and_transform(state, "STATEFP"):
        state.plot(color="0.9", ax=ax, edgecolor="0.8", linewidth=0.5)
    ax.set_xlim(plt.xlim())
    ax.set_ylim(plt.ylim())
    return fig, ax


def _project_and_transform(data: gpd.GeoDataFrame,
                           state_col: str = "state_fips") -> Tuple:
    """Segment and break GeoDF into Contiguous US, Alaska, Puerto Rico, and Hawaii for plotting.

    Given GeoDF with state fips column, break into Contiguous US, Alaska, Puerto Rico, and Hawaii
    GeoDFs with their own Albers Equal Area Conic Projections.

    Also scales and translates so Alaska and Hawaii are in the bottom left corner and Puerto Rico
    is closer to Hawaii.

    :param data: GeoDF with shape info and a column designating the state.
    :param state_col: Name of column with state FIPS codes.
    :return: Tuple of four GeoDFs: Contiguous US, Alaska, Hawaii, and Puerto Rico.
    """
    cont = data.loc[[i in CONTIGUOUS_FIPS for i in data[state_col]], :].to_crs("ESRI:102003")
    alaska = data.loc[data[state_col] == "02", :].to_crs("ESRI:102006")
    pr = data.loc[data[state_col] == "72", :].to_crs("ESRI:102003")
    hawaii = data.loc[data[state_col] == "15", :].to_crs("ESRI:102007")

    alaska.geometry = alaska.geometry.scale(0.35, 0.35, origin=(0, 0)).translate(-1.8e6, -1.6e6)
    hawaii.geometry = hawaii.geometry.translate(-1e6, -2e6)
    pr.geometry = pr.geometry.translate(-0.9e6, 1e6).rotate(-16, origin=(0, 0))
    return cont, alaska, pr, hawaii


def _join_state_geo_df(data: pd.DataFrame,
                       state_col: str,
                       geo_info: gpd.GeoDataFrame,
                       join_type: str = "right") -> gpd.GeoDataFrame:
    """Join DF information to polygon information in a GeoDF at the state level.

    :param data: DF with state info
    :param state_col: name of column in `data` containing state info to join on
    :param geo_info: GeoDF of state shape info read from Census shapefiles
    :param join_type: Type of join to do between input data (left side) and geo data (right side).
      Must be one of {‘left’, ‘right’, ‘outer’, ‘inner’}.
    :return: ``data`` with state polygon and state FIPS joined.
    """
    input_cols = list(data.columns)
    geo_info.STUSPS = [i.lower() for i in geo_info.STUSPS]  # lowercase for consistency
    merged = data.merge(geo_info, how=join_type, left_on=state_col, right_on="STUSPS", sort=True)
    # use full state list in the return
    merged[state_col] = merged.STUSPS.combine_first(merged[state_col])
    merged.rename(columns={"STATEFP": "state_fips"}, inplace=True)
    return gpd.GeoDataFrame(merged[input_cols + ["state_fips", "geometry"]])


def _join_county_geo_df(data: pd.DataFrame,
                        county_col: str,
                        geo_info: gpd.GeoDataFrame,
                        join_type: str = "right",
                        combine_megacounties: bool = False) -> gpd.GeoDataFrame:
    """Join DF information to polygon information in a GeoDF at the county level.

    Counties with no direct key in the data DF will have the megacounty value joined.

    :param data: DF with county info.
    :param county_col: name of column in `data` containing county info to join on.
    :param geo_info: GeoDF of county shape info read from Census shapefiles.
    :param join_type: Type of join to do between input data (left side) and geo data (right side).
      Must be one of {‘left’, ‘right’, ‘outer’, ‘inner’}.
    :param combine_megacounties: For each state, return all counties without a signal value as a
      single polygon with the megacounty value.
    :return: ``data`` with county polygon and state fips joined.
    """
    input_cols = list(data.columns)
    # create state FIPS code in copy, otherwise original gets modified
    data = data.assign(state=[i[:2] for i in data[county_col]])
    if combine_megacounties:
        merged = _combine_megacounties(data, county_col, geo_info)
    else:
        merged = _distribute_megacounties(data, county_col, geo_info, join_type)
    merged[county_col] = merged.GEOID.combine_first(merged[county_col])
    merged.rename(columns={"STATEFP": "state_fips"}, inplace=True)
    return gpd.GeoDataFrame(merged[input_cols + ["state_fips", "geometry"]])


def _combine_megacounties(data: pd.DataFrame,
                          county_col: str,
                          geo_info: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Join a DataFrame of county signals with a GeoDataFrame of polygons for plotting.

    Merges a DataFrame of counties and signals with a DataFrame of county polygons. Megacounties,
    if present, are assigned a polygon which is the union of all counties in the state with no
    signal value.

    :param data: DataFrame of county signals.
    :param county_col: Name of column containing county.
    :parm geo_info: GeoDataFrame of counties and corresponding polygons.
    :return: ``data`` with county polygon and state fips joined. No polgyon information is
      provided for counties without a signal value since they are captured by the megacounty
      polygon.
    """
    merged = data.merge(geo_info, how="left", left_on=county_col, right_on="GEOID", sort=True)
    missing = set(geo_info.GEOID) - set(data[county_col])
    for i, row in merged.iterrows():
        if _is_megacounty(row[county_col]):
            state = row[county_col][:2]
            state_missing = [j for j in missing if j.startswith(state)]
            combined_poly = geo_info.loc[geo_info.GEOID.isin(state_missing), "geometry"].unary_union
            # pandas has a bug when assigning MultiPolygons, so you need to do this weird workaround
            # https://github.com/geopandas/geopandas/issues/992
            merged.loc[[i], "geometry"] = gpd.GeoSeries(combined_poly).values
            merged.loc[[i], "STATEFP"] = state
    return merged


def _distribute_megacounties(data: pd.DataFrame,
                             county_col: str,
                             geo_info: gpd.GeoDataFrame,
                             join_type: str = "right") -> gpd.GeoDataFrame:
    """Join a DataFrame of county signals with a GeoDataFrame of polygons for plotting.

    Merges a DataFrame of counties and signals with a DataFrame of county polygons. Counties
    without a value but with a corresponding megacounty take on the megacounty value.

    :param data: DataFrame of county signals.
    :param county_col: Name of column containing county.
    :param geo_info: GeoDataFrame of counties and corresponding polygons.
    :param join_type: Type of join to do between input data (left side) and geo data (right side).
      Must be one of {‘left’, ‘right’, ‘outer’, ‘inner’}.
    :return: ``data`` with county polygon and state fips joined. No polgyon information is
      provided for megacounties.
    """
    # join all counties with valid FIPS
    merged = data.merge(geo_info, how=join_type, left_on=county_col, right_on="GEOID", sort=True)
    mega_df = data.loc[[_is_megacounty(i) for i in data[county_col]], :]
    if not mega_df.empty and join_type == "right":
        # if mega counties exist, join them on state
        merged = merged.merge(mega_df, how="left", left_on="STATEFP", right_on="state", sort=True)
        # if no county value present, us the megacounty values
        for c in data.columns:
            merged[c] = merged[f"{c}_x"].combine_first(merged[f"{c}_y"])
    return merged


def _join_msa_geo_df(data: pd.DataFrame,
                     msa_col: str,
                     geo_info: gpd.GeoDataFrame,
                     join_type: str = "right") -> gpd.GeoDataFrame:
    """Join DF information to polygon information in a GeoDF at the MSA level.

    For MSAs which span multiple states, the first state in the name is returned for the state FIPS.

    :param data: DF with state info
    :param msa_col: cname of column in `data` containing state info to join on
    :param geo_info: GeoDF of state shape info read from Census shapefiles
    :param join_type: Type of join to do between input data (left side) and geo data (right side).
      Must be one of {‘left’, ‘right’, ‘outer’, ‘inner’}.
    :return: ``data`` with cbsa polygon and state fips joined.
    """
    geo_info = geo_info[geo_info.LSAD == "M1"]  # only get metro and not micropolitan areas
    input_cols = list(data.columns)
    merged = data.merge(geo_info, how=join_type, left_on=msa_col, right_on="GEOID", sort=True)
    # use full state list in the return
    merged[msa_col] = merged.GEOID.combine_first(merged[msa_col])
    # get the first state, which will be the first two characters after the comma and whitespace
    merged["state_fips"] = [STATE_ABBR_TO_FIPS.get(i.split(",")[1][1:3]) for i in merged.NAME]
    return gpd.GeoDataFrame(merged[input_cols + ["state_fips", "geometry"]])


def _join_hrr_geo_df(data: pd.DataFrame,
                     msa_col: str,
                     geo_info: gpd.GeoDataFrame,
                     join_type: str = "right") -> gpd.GeoDataFrame:
    """Join DF information to polygon information in a GeoDF at the HRR level.

    :param data: DF with state info
    :param msa_col: cname of column in `data` containing state info to join on
    :param geo_info: GeoDF of state shape info read from Census shapefiles
    :param join_type: Type of join to do between input data (left side) and geo data (right side).
      Must be one of {‘left’, ‘right’, ‘outer’, ‘inner’}.
    :return: ``data`` with HRR polygon and state fips joined.
    """
    geo_info["hrr_num"] = geo_info.hrr_num.astype("int").astype(str)  # original col was a float
    input_cols = list(data.columns)
    merged = data.merge(geo_info, how=join_type, left_on=msa_col, right_on="hrr_num", sort=True)
    # use full state list in the return
    merged[msa_col] = merged.hrr_num.combine_first(merged[msa_col])
    # get the first state, which will be the first two characters in the HRR name
    merged["state_fips"] = [STATE_ABBR_TO_FIPS.get(i[:2]) for i in merged.hrr_name]
    return gpd.GeoDataFrame(merged[input_cols + ["state_fips", "geometry"]])


def _is_megacounty(fips: str) -> bool:
    """Determine if a code is a megacounty.

    :param fips: FIPS code to test.
    :return: Boolean for if the input code is a megacounty or not.

    """
    return fips.endswith("000") and len(fips) == 5



"""This is the client side library for accessing the COVIDcast API."""
import warnings
from datetime import timedelta, date
from functools import reduce
from typing import Union, Iterable, Tuple, List

import pandas as pd
from delphi_epidata import Epidata

from .errors import NoDataWarning

# Point API requests to the AWS endpoint
Epidata.BASE_URL = "https://api.covidcast.cmu.edu/epidata/api.php"

VALID_GEO_TYPES = {"county", "hrr", "msa", "dma", "state"}


def signal(data_source: str,
           signal: str,  # pylint: disable=W0621
           start_day: date = None,
           end_day: date = None,
           geo_type: str = "county",
           geo_values: Union[str, Iterable[str]] = "*",
           as_of: date = None,
           issues: Union[date, Tuple[date], List[date]] = None,
           lag: int = None) -> Union[pd.DataFrame, None]:
    """Download a Pandas data frame for one signal.

    Obtains data for selected date ranges for all geographic regions of the
    United States. Available data sources and signals are documented in the
    `COVIDcast signal documentation
    <https://cmu-delphi.github.io/delphi-epidata/api/covidcast_signals.html>`_.
    Most (but not all) data sources are available at the county level, but the
    API can also return data aggregated to metropolitan statistical areas,
    hospital referral regions, or states, as desired, by using the ``geo_type``
    argument.

    The COVIDcast API tracks updates and changes to its underlying data, and
    records the first date each observation became available. For example, a
    data source may report its estimate for a specific state on June 3rd on June
    5th, once records become available. This data is considered "issued" on June
    5th. Later, the data source may update its estimate for June 3rd based on
    revised data, creating a new issue on June 8th. By default, ``signal()``
    returns the most recent issue available for every observation. The
    ``as_of``, ``issues``, and ``lag`` parameters allow the user to select
    specific issues instead, or to see all updates to observations. These
    options are mutually exclusive; if you specify more than one, ``as_of`` will
    take priority over ``issues``, which will take priority over ``lag``.

    Note that the API only tracks the initial value of an estimate and *changes*
    to that value. If a value was first issued on June 5th and never updated,
    asking for data issued on June 6th (using ``issues`` or ``lag``) would *not*
    return that value, though asking for data ``as_of`` June 6th would.

    Note also that the API enforces a maximum result row limit; results beyond
    the maximum limit are truncated. This limit is sufficient to fetch
    observations in all counties in the United States on one day. This client
    automatically splits queries for multiple days across multiple API calls.
    However, if data for one day has been issued many times, using the
    ``issues`` argument may return more results than the query limit. A warning
    will be issued in this case. To see all results, split your query across
    multiple calls with different ``issues`` arguments.

    See the `COVIDcast API documentation
    <https://cmu-delphi.github.io/delphi-epidata/api/covidcast.html>`_ for more
    information on available geography types, signals, and data formats, and
    further discussion of issue dates and data versioning.

    :param data_source: String identifying the data source to query, such as
      ``"fb-survey"``.
    :param signal: String identifying the signal from that source to query,
      such as ``"smoothed_cli"``.
    :param start_day: Query data beginning on this date. Provided as a
      ``datetime.date`` object. If ``start_day`` is ``None``, defaults to the
      first day data is available for this signal.
    :param end_day: Query data up to this date, inclusive. Provided as a
      ``datetime.date`` object. If ``end_day`` is ``None``, defaults to the most
      recent day data is available for this signal.
    :param geo_type: The geography type for which to request this data, such as
      ``"county"`` or ``"state"``. Available types are described in the
      COVIDcast signal documentation. Defaults to ``"county"``.
    :param geo_values: The geographies to fetch data for. The default, ``"*"``,
      fetches all geographies. To fetch one geography, specify its ID as a
      string; multiple geographies can be provided as an iterable (list, tuple,
      ...) of strings.
    :param as_of: Fetch only data that was available on or before this date,
      provided as a ``datetime.date`` object. If ``None``, the default, return
      the most recent available data.
    :param issues: Fetch only data that was published or updated ("issued") on
      these dates. Provided as either a single ``datetime.date`` object,
      indicating a single date to fetch data issued on, or a tuple or list
      specifying (start, end) dates. In this case, return all data issued in
      this range. There may be multiple rows for each observation, indicating
      several updates to its value. If ``None``, the default, return the most
      recently issued data.
    :param lag: Integer. If, for example, ``lag=3``, fetch only data that was
      published or updated exactly 3 days after the date. For example, a row
      with ``time_value`` of June 3 will only be included in the results if its
      data was issued or updated on June 6. If ``None``, the default, return the
      most recently issued data regardless of its lag.
    :returns: A Pandas data frame with matching data. Each row is one
      observation on one day in one geographic location. Contains the following
      columns:

      ``geo_value``
        Identifies the location, such as a state name or county FIPS code. The
        geographic coding used by COVIDcast is described in the `API
        documentation here
        <https://cmu-delphi.github.io/delphi-epidata/api/covidcast_geography.html>`_.

      ``signal``
        Name of the signal, same as the value of the ``signal`` input argument. Used for
        downstream functions to recognize where this signal is from.

      ``time_value``
        Contains a `pandas Timestamp object
        <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Timestamp.html>`_
        identifying the date this estimate is for.

      ``issue``
        Contains a `pandas Timestamp object
        <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Timestamp.html>`_
        identifying the date this estimate was issued. For example, an estimate
        with a ``time_value`` of June 3 might have been issued on June 5, after
        the data for June 3rd was collected and ingested into the API.

      ``lag``
        Integer giving the difference between ``issue`` and ``time_value``,
        in days.

      ``value``
        The signal quantity requested. For example, in a query for the
        ``confirmed_cumulative_num`` signal from the ``usa-facts`` source,
        this would be the cumulative number of confirmed cases in the area, as
        of the ``time_value``.

      ``stderr``
        The value's standard error, if available.

      ``sample_size``
        Indicates the sample size available in that geography on that day;
        sample size may not be available for all signals, due to privacy or
        other constraints.

      ``geo_type``
        Geography type for the signal, same as the value of the ``geo_type`` input argument.
        Used for downstream functions to parse ``geo_value`` correctly

      ``data_source``
        Name of the signal source, same as the value of the ``data_source`` input argument. Used for
        downstream functions to recognize where this signal is from.

    Consult the `signal documentation
    <https://cmu-delphi.github.io/delphi-epidata/api/covidcast_signals.html>`_
    for more details on how values and standard errors are calculated for
    specific signals.

    """
    if geo_type not in VALID_GEO_TYPES:
        raise ValueError("geo_type must be one of " + ", ".join(VALID_GEO_TYPES))

    if start_day is None or end_day is None:
        signal_meta = _signal_metadata(data_source, signal, geo_type)

    start_day = signal_meta["min_time"].to_pydatetime().date() \
        if start_day is None else start_day

    end_day = signal_meta["max_time"].to_pydatetime().date() \
        if end_day is None else end_day

    if start_day > end_day:
        raise ValueError("end_day must be on or after start_day, but "
                         "start_day = '{start}', end_day = '{end}'".format(
                             start=start_day, end=end_day))

    if isinstance(geo_values, str):
        # User only provided one, not a list
        geo_values = [geo_values]

    dfs = [
        _fetch_single_geo(
            data_source, signal, start_day, end_day, geo_type, geo_value,
            as_of, issues, lag)
        for geo_value in set(geo_values)
    ]

    try:
        # pd.concat automatically filters out None
        out = pd.concat(dfs)
    except ValueError:
        # pd.concat raises ValueError if all of the dfs are None, meaning we
        # found no data
        return None

    return out


def metadata() -> pd.DataFrame:
    """Fetch COVIDcast surveillance stream metadata.

    Obtains a data frame of metadata describing all publicly available data
    streams from the COVIDcast API. See the `data source and signals
    documentation
    <https://cmu-delphi.github.io/delphi-epidata/api/covidcast_signals.html>`_
    for descriptions of the available sources.

    :returns: A data frame containing one row per available signal, with the
      following columns:

      ``data_source``
        Data source name.

      ``signal``
        Signal name.

      ``time_type``
        Temporal resolution at which this signal is reported. "day", for
        example, means the signal is reported daily.

      ``geo_type``
        Geographic level for which this signal is available, such as county,
        state, msa, or hrr. Most signals are available at multiple geographic
        levels and will hence be listed in multiple rows with their own
        metadata.

      ``min_time``
        First day for which this signal is available.

      ``max_time``
        Most recent day for which this signal is available.

      ``num_locations``
        Number of distinct geographic locations available for this signal. For
        example, if `geo_type` is county, the number of counties for which this
        signal has ever been reported.

      ``min_value``
        The smallest value that has ever been reported.

      ``max_value``
        The largest value that has ever been reported.

      ``mean_value``
        The arithmetic mean of all reported values.

      ``stdev_value``
        The sample standard deviation of all reported values.

      ``last_update``
        The UTC datetime for when the signal value was last updated.

      ``max_issue``
        Most recent date data was issued.

      ``min_lag``
        Smallest lag from observation to issue, in days.

      ``max_lag``
        Largest lag from observation to issue, in days.
    """
    meta = Epidata.covidcast_meta()

    if meta["result"] != 1:
        # Something failed in the API and we did not get real metadata
        raise RuntimeError("Error when fetching metadata from the API",
                           meta["message"])

    meta_df = pd.DataFrame.from_dict(meta["epidata"])
    meta_df["min_time"] = pd.to_datetime(meta_df["min_time"], format="%Y%m%d")
    meta_df["max_time"] = pd.to_datetime(meta_df["max_time"], format="%Y%m%d")
    meta_df["last_update"] = pd.to_datetime(meta_df["last_update"], unit="s")
    return meta_df


def aggregate_signals(signals: list, dt: list = None, join_type: str = "outer") -> pd.DataFrame:
    """Given a list of DataFrames, [optionally] lag each one and join them into one DataFrame.

    This method takes a list of DataFrames containing signal information for
    geographic regions across time, and outputs a single DataFrame with a column
    for each signal value for each region/time. The ``data_source``,
    ``signal``, and index of each DataFrame in ``signals`` are appended to the
    front of each output column name separated by underscores (e.g.
    ``source_signal_0_inputcolumn``), and the original data_source and signal
    columns will be dropped. The input DataFrames must all be of the same
    geography type, and a single ``geo_type`` column will be returned in the final
    DataFrame.

    Each signal's time value can be shifted for analysis on lagged signals using the ``dt``
    argument, which takes a list of integer days to lag each signal's date. Lagging a signal by +1
    day means that all the dates get shifted forward by 1 day (e.g. Jan 1 becomes Jan 2).

    :param signals: List of DataFrames to join.
    :param dt: List of lags in days for each of the input DataFrames in ``signals``.
      Defaults to ``None``. When provided, must be the same length as ``signals``.
    :param join_type: Type of join to be done between the DataFrames in ``signals``.
      Defaults to ``"outer"``, so the output DataFrame contains all region/time
      combinations at which at least one signal was observed.
    :return: DataFrame of aggregated signals.

    """
    if dt is not None and len(dt) != len(signals):
        raise ValueError("Length of `dt` must be same as length of `signals`")
    dt = [0] * len(signals) if not dt else dt
    join_cols = ["time_value", "geo_value"]
    dt_dfs = []
    first_geo_type = _detect_metadata(signals[0])[2]

    for i, (df, lag) in enumerate(zip(signals, dt)):
        df_c = df.copy()  # make a copy so we don't modify originals
        source, sig_type, geo_type = _detect_metadata(df_c)
        if geo_type != first_geo_type:
            raise ValueError("Multiple geo_types detected. "
                             "All signals must have the same geo_type to be aggregated.")

        df_c["time_value"] = [day + timedelta(lag) for day in df_c["time_value"]]  # lag dates
        df_c.drop(["signal", "data_source", "geo_type"], axis=1, inplace=True)
        df_c.rename(
            columns={j: f"{source}_{sig_type}_{i}_{j}" for j in df_c.columns if j not in join_cols},
            inplace=True)
        dt_dfs.append(df_c)

    joined_df = reduce(lambda x, y: pd.merge(x, y, on=join_cols, how=join_type, sort=True), dt_dfs)
    joined_df["geo_type"] = geo_type
    return joined_df


def _detect_metadata(data: pd.DataFrame,
                     data_source_col: str = "data_source",
                     signal_col: str = "signal",
                     geo_type_col: str = "geo_type") -> Tuple:
    """Given a DataFrame, return the signal attributes of that DataFrame.

    Raises ValueError if any of the columns are heterogeneous.
    Inputs must have all three of the relevant columns: data source, signal, and geography type.

    :param data: DataFrame with data_source, signal, and geo_tye columns.
    :param data_source_col: name of column with data source info
    :param signal_col: name of column with signal info
    :param geo_type_col: name of column with geography type
    :return: tuple of the three types
    """
    unique_data_source_vals = data[data_source_col].unique()
    unique_signal_col_vals = data[signal_col].unique()
    unique_geo_type_vals = data[geo_type_col].unique()
    if len(unique_data_source_vals) > 1:
        raise ValueError("Multiple data sources detected.")
    if len(unique_signal_col_vals) > 1:
        raise ValueError("Multiple signals detected.")
    if len(unique_geo_type_vals) > 1:
        raise ValueError("Multiple geography types detected.")
    return unique_data_source_vals[0], unique_signal_col_vals[0], unique_geo_type_vals[0]


def _fetch_single_geo(data_source: str,
                      signal: str,  # pylint: disable=W0621
                      start_day: date,
                      end_day: date,
                      geo_type: str,
                      geo_value: str,
                      as_of: date,
                      issues: Union[date, tuple, list],
                      lag: int) -> Union[pd.DataFrame, None]:
    """Fetch data for a single geo.

    signal() wraps this to support fetching data over an iterable of
    geographies, and stacks the resulting data frames.

    If no data is found, return None, so signal() can easily filter out these
    entries.

    """
    as_of_str = _date_to_api_string(as_of) if as_of is not None else None
    issues_strs = _dates_to_api_strings(issues) if issues is not None else None

    cur_day = start_day

    dfs = []

    while cur_day <= end_day:
        day_str = _date_to_api_string(cur_day)

        day_data = Epidata.covidcast(data_source, signal, time_type="day",
                                     geo_type=geo_type, time_values=day_str,
                                     geo_value=geo_value, as_of=as_of_str,
                                     issues=issues_strs, lag=lag)

        # Two possible error conditions: no data or too much data.
        if day_data["message"] == "no results":
            warnings.warn(f"No {data_source} {signal} data found on {day_str} "
                          f"for geography '{geo_type}'",
                          NoDataWarning)
        if day_data["message"] not in {"success", "no results"}:
            warnings.warn(f"Problem obtaining {data_source} {signal} data on {day_str} "
                          f"for geography '{geo_type}': {day_data['message']}",
                          RuntimeWarning)

        # In the too-much-data case, we continue to try putting the truncated
        # data in our results. In the no-data case, skip this day entirely,
        # since there is no "epidata" in the response.
        if "epidata" in day_data:
            dfs.append(pd.DataFrame.from_dict(day_data["epidata"]))

        cur_day += timedelta(1)

    if len(dfs) > 0:
        out = pd.concat(dfs)
        out.drop("direction", axis=1, inplace=True)
        out["time_value"] = pd.to_datetime(out["time_value"], format="%Y%m%d")
        out["issue"] = pd.to_datetime(out["issue"], format="%Y%m%d")
        out["geo_type"] = geo_type
        out["data_source"] = data_source
        out["signal"] = signal
        return out

    return None


def _signal_metadata(data_source: str,
                     signal: str,  # pylint: disable=W0621
                     geo_type: str) -> dict:
    """Fetch metadata for a single signal as a dict."""
    meta = metadata()

    mask = ((meta.data_source == data_source) &
            (meta.signal == signal) &
            (meta.time_type == "day") &
            (meta.geo_type == geo_type))

    matches = meta[mask]

    if matches.shape[0] == 0:
        raise ValueError("Unable to find metadata for source '{source}', "
                         "signal '{signal}', at '{geo_type}' "
                         "resolution.".format(
                             source=data_source,
                             signal=signal,
                             geo_type=geo_type))

    assert matches.shape[0] == 1, "it should be impossible to have two identical signals"
    output: dict = matches.to_dict("records")[0]
    return output


def _date_to_api_string(date: date) -> str:  # pylint: disable=W0621
    """Convert a date object to a YYYYMMDD string expected by the API."""
    return date.strftime("%Y%m%d")


def _dates_to_api_strings(dates: Union[date, list, tuple]) -> str:
    """Convert a date object, or pair of (start, end) objects, to YYYYMMDD strings."""
    if not isinstance(dates, (list, tuple)):
        return _date_to_api_string(dates)

    return "-".join(_date_to_api_string(date) for date in dates)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="covidcast",
    version="0.1.1",  # also update in docs/conf.py
    author="Alex Reinhart",
    author_email="areinhar@stat.cmu.edu",
    description="Access COVID-19 data through the Delphi COVIDcast API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://cmu-delphi.github.io/covidcast/covidcast-py/html/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "pandas",
        "requests",
        "delphi-epidata>=0.0.7",
        "geopandas",
        "matplotlib",
        "numpy",
        "descartes",
        "imageio-ffmpeg",
        "imageio",
        "tqdm"
    ],
    package_data={"covidcast": ["shapefiles/*/*", "geo_mappings/*"]}
)
